@@ -21,9 +21,14 @@ import os
 import json
 import logging
 import shutil
+import struct
 from builtins import str
 from builtins import bytes
 from future.utils import iteritems, iterkeys, itervalues
+from concurrent.futures import ThreadPoolExecutor, wait, ALL_COMPLETED, FIRST_COMPLETED
+import threading
+import functools
+import multiprocessing
 
 import baidubce
 from baidubce import bce_client_configuration
@@ -44,7 +49,6 @@ from baidubce.services.bos import storage_class
 from baidubce.utils import required
 from baidubce import compat
 
-
 _logger = logging.getLogger(__name__)
 
 FETCH_MODE_SYNC = b"sync"
@@ -53,6 +57,32 @@ FETCH_MODE_ASYNC = b"async"
 ENCRYPTION_ALGORITHM= "AES256"
 
 
+class UploadTaskHandle:
+    """
+    handle to control multi upload file with multi-thread
+    """
+    def __init__(self):
+        self.cancel_flag= False
+        self.cancel_lock = threading.Lock()
+
+    def cancel(self):
+        """
+        cancel putting super object from file with multi-thread
+        """
+        self.cancel_lock.acquire()
+        self.cancel_flag= True
+        self.cancel_lock.release()
+
+    def is_cancel(self):
+        """
+        get cancel flag
+        """
+        self.cancel_lock.acquire()
+        result = self.cancel_flag
+        self.cancel_lock.release()
+        return result
+
+
 class BosClient(BceBaseClient):
     """
     sdk client
@@ -479,6 +509,76 @@ class BosClient(BceBaseClient):
                            params={b'replicationProgress': b''},
                            config=config)
 
+    @required(bucket_name=(bytes, str), inventory=(dict))
+    def put_bucket_inventory(self, bucket_name, inventory, config=None):
+        """
+        set bucket inventoru
+
+        :type bucket: string
+        :param bucket: None
+
+        :type inventory: dict
+        :param inventory: configuration for bucket inventory
+
+        :return:
+            **HttpResponse Class**
+        """
+        conf_id = compat.convert_to_bytes(inventory["id"])
+        return self._send_request(http_methods.PUT,
+                           bucket_name,
+                           body=json.dumps(inventory,
+                                           default=BosClient._dump_acl_object),
+                           headers={http_headers.CONTENT_TYPE: http_content_types.JSON},
+                           params={b'inventory': b'', b'id': conf_id},
+                           config=config)
+
+    @required(bucket_name=(bytes, str), inventory_conf_id=(bytes, str))
+    def get_bucket_inventory(self, bucket_name, inventory_conf_id, config=None):
+        """
+        Get configuration of bucket inventory
+
+        :type bucket: string
+        :param bucket: None
+
+        :return:
+            **HttpResponse Class**
+        """
+        return self._send_request(http_methods.GET,
+                           bucket_name,
+                           params={b'inventory': b'', b'id': compat.convert_to_bytes(inventory_conf_id)},
+                           config=config)
+
+    @required(bucket_name=(bytes, str), inventory_conf_id=(bytes, str))
+    def delete_bucket_inventory(self, bucket_name, inventory_conf_id, config=None):
+        """
+        Delete configuration of bucket inventory
+
+        :type bucket: string
+        :param bucket: None
+
+        :return:
+            **HttpResponse Class**
+        """
+        return self._send_request(http_methods.DELETE,
+                           bucket_name,
+                           params={b'inventory': b'', b'id': compat.convert_to_bytes(inventory_conf_id)},
+                           config=config)
+
+    @required(bucket_name=(bytes, str))
+    def list_bucket_inventory(self, bucket_name, config=None):
+        """
+        list configuration of bucket inventory
+
+        :type bucket: string
+        :param bucket: None
+
+        :return:
+            **HttpResponse Class**
+        """
+        return self._send_request(http_methods.GET,
+                           bucket_name,
+                           params={b'inventory': b''},
+                           config=config)
 
     @required(bucket_name=(bytes, str))
     def put_bucket_trash(self, bucket_name, trash_dir=None, config=None):
@@ -547,9 +647,11 @@ class BosClient(BceBaseClient):
                                 params=None,
                                 headers_to_sign=None,
                                 protocol=None,
-                                config=None):
+                                config=None,
+                                httpmethod=http_methods.GET):
         """
-        Get an authorization url with expire time
+        Get an authorization url with expire time.
+        specified  protocol in endpoint > protocal > default protocol in config.
 
         :type timestamp: int
         :param timestamp: None
@@ -567,28 +669,31 @@ class BosClient(BceBaseClient):
         config = self._merge_config(config)
         headers = headers or {}
         params = params or {}
-        # specified protocol > protocol in endpoint > default protocol
+
+        # specified  protocol in endpoint > protocal > default protocol in config
+        if protocol is not None:
+            config.protocol = protocol
         endpoint_protocol, endpoint_host, endpoint_port = \
             utils.parse_host_port(config.endpoint, config.protocol)
-        protocol = protocol or endpoint_protocol
 
         full_host = endpoint_host
-        if endpoint_port != config.protocol.default_port:
+        if endpoint_port != endpoint_protocol.default_port:
             full_host += b':' + compat.convert_to_bytes(endpoint_port)
         headers[http_headers.HOST] = full_host
 
         path = self._get_path(config, bucket_name, key)
-
+        if httpmethod != http_methods.GET and httpmethod != http_methods.HEAD:
+            headers_to_sign = set([b'host'])
         params[http_headers.AUTHORIZATION.lower()] = bce_v1_signer.sign(
             config.credentials,
-            http_methods.GET,
+            httpmethod,
             path,
             headers,
             params,
             timestamp,
             expiration_in_seconds,
             headers_to_sign)
-        return b"%s://%s%s?%s" % (compat.convert_to_bytes(protocol.name),
+        return b"%s://%s%s?%s" % (compat.convert_to_bytes(endpoint_protocol.name),
                                  full_host,
                                  path,
                                  utils.get_canonical_querystring(params, False))
@@ -854,6 +959,11 @@ class BosClient(BceBaseClient):
             f.close()
         return True
 
+    @staticmethod
+    def _parse_select_message(http_response, response, select_response):
+        select_response.init_from_http_response(http_response, response)
+        return True
+
     @required(bucket_name=(bytes, str), key=(bytes, str))
     def get_object_as_string(self, bucket_name, key, range=None, config=None):
         """
@@ -1032,6 +1142,9 @@ class BosClient(BceBaseClient):
                    user_metadata=None,
                    storage_class=None,
                    user_headers=None,
+                   encryption=None,
+                   customer_key=None,
+                   customer_key_md5=None,
                    config=None):
         """
         Put object and put content of file to the object
@@ -1084,6 +1197,9 @@ class BosClient(BceBaseClient):
                                user_metadata=None,
                                storage_class=None,
                                user_headers=None,
+                               encryption=None,
+                               customer_key=None,
+                               customer_key_md5=None,
                                config=None):
         """
         Create object and put content of string to the object
@@ -1120,6 +1236,9 @@ class BosClient(BceBaseClient):
                                    user_metadata=user_metadata,
                                    storage_class=storage_class,
                                    user_headers=user_headers,
+                                   encryption=encryption,
+                                   customer_key=customer_key,
+                                   customer_key_md5=customer_key_md5,
                                    config=config)
         finally:
             if fp is not None:
@@ -1134,6 +1253,9 @@ class BosClient(BceBaseClient):
                              user_metadata=None,
                              storage_class=None,
                              user_headers=None,
+                             encryption=None,
+                             customer_key=None,
+                             customer_key_md5=None,
                              config=None):
 
         """
@@ -1174,6 +1296,9 @@ class BosClient(BceBaseClient):
                                    user_metadata=user_metadata,
                                    storage_class=storage_class,
                                    user_headers=user_headers,
+                                   encryption=encryption,
+                                   customer_key=customer_key,
+                                   customer_key_md5=customer_key_md5,
                                    config=config)
         finally:
             fp.close()
@@ -1709,6 +1834,88 @@ class BosClient(BceBaseClient):
             else:
                 break
 
+    def _upload_task(self, bucket_name, object_key, upload_id,
+        part_number, part_size, file_name, offset, part_list, uploadTaskHandle):
+        if uploadTaskHandle.is_cancel():
+            _logger.debug("upload task canceled with partNumber={}!".format(part_number))
+            return
+        try:
+            response = self.upload_part_from_file(bucket_name, object_key, upload_id,
+                part_number, part_size, file_name, offset)
+            part_list.append({
+                "partNumber": part_number,
+                "eTag": response.metadata.etag
+            })
+            _logger.debug("upload task success with partNumber={}!".format(part_number))
+        except Exception as e:
+            _logger.debug("upload task failed with partNumber={}!".format(part_number))
+            #_logger.debug(e)
+
+    @required(bucket_name=(bytes, str), key=(bytes, str), file_name=(bytes, str))
+    def put_super_obejct_from_file(self, bucket_name, key, file_name, chunk_size=5,
+            thread_num=None,
+            uploadTaskHandle=None,
+            content_type=None,
+            storage_class=None,
+            user_headers=None,
+            config=None):
+        """
+        Multipart Upload file to bos
+
+        param chunk_size: part size , default part size is 5MB
+        """
+        # check params
+        if chunk_size > 5 * 1024 or chunk_size <= 0:
+           raise BceClientError("chunk size is valid, it should be more than 0 and not nore than 5120!")
+        left_size = os.path.getsize(file_name)
+        # if file size more than 5TB, reject
+        if left_size > 5 * 1024 * 1024 * 1024 * 1024:
+           raise BceClientError("File size must not be more than 5TB!")
+        if thread_num is None or thread_num <= 1:
+           thread_num = multiprocessing.cpu_count()
+        part_size = chunk_size * 1024 * 1024
+        total_part = left_size // part_size
+        if left_size % part_size != 0:
+            total_part += 1
+        if uploadTaskHandle is None:
+            uploadTaskHandle = UploadTaskHandle()
+        # initial
+        upload_id = self.initiate_multipart_upload(bucket_name, key,
+                content_type=content_type,
+                storage_class=storage_class,
+                user_headers=user_headers).upload_id
+
+        executor = ThreadPoolExecutor(thread_num)
+        all_tasks = []
+        offset = 0
+        part_number = 1
+        part_list = []
+
+        while left_size > 0:
+            if left_size < part_size:
+                part_size = left_size
+            temp_task= executor.submit(self._upload_task, bucket_name, key, upload_id, part_number, part_size,
+                file_name, offset, part_list, uploadTaskHandle)
+            all_tasks.append(temp_task)
+            left_size -= part_size
+            offset += part_size
+            part_number += 1
+        # wait all upload task to exit
+        wait(all_tasks, return_when=ALL_COMPLETED)
+        if uploadTaskHandle.is_cancel():
+            _logger.debug("putting super object is canceled!")
+            self.abort_multipart_upload(bucket_name, key, upload_id = upload_id)
+            return False
+        elif len(part_list) != total_part:
+            _logger.debug("putting super object failed!")
+            self.abort_multipart_upload(bucket_name, key, upload_id = upload_id)
+            return False
+        # sort
+        part_list.sort(key=lambda x: x["partNumber"])
+        # complete_multipart_upload
+        self.complete_multipart_upload(bucket_name, key, upload_id, part_list)
+        return True
+
     @required(bucket_name=(bytes, str), key=(bytes, str), acl=(list, dict))
     def set_object_acl(self, bucket_name, key, acl, config=None):
         """
@@ -1870,6 +2077,102 @@ class BosClient(BceBaseClient):
                 params={b'fetch': b''},
                 config=config)
 
+    @required(bucket_name=(bytes, str), key=(bytes, str), symlink=(bytes, str), forbid_overwrite=(bool))
+    def put_object_symlink(self, bucket_name, key, symlink, forbid_overwrite=None, 
+            user_metadata=None,
+            storage_class=None,
+            config=None):
+        """
+        put object symlink
+
+        :type bucket: string
+        :param bucket: None
+
+        :type key: string
+        :type key: object name
+
+        :type symlink: string
+        :type symlink_key: symlink name
+
+        :return:
+            **HttpResponse Class**
+        """
+        key = compat.convert_to_bytes(key)
+        symlink = compat.convert_to_bytes(symlink)
+        headers = self._prepare_object_headers(user_metadata=user_metadata,
+                storage_class=storage_class)
+        headers[http_headers.BOS_SYMLINK_TARGET] = key
+        if forbid_overwrite is not None:
+            if forbid_overwrite:
+                headers[http_headers.BOS_FORBID_OVERWRITE] = b'true'
+            else:
+                headers[http_headers.BOS_FORBID_OVERWRITE] = b'false'
+        return self._send_request(http_methods.PUT,
+                           bucket_name,
+                           symlink,
+                           headers=headers,
+                           params={b'symlink': b''},
+                           config=config)
+
+
+    @required(bucket_name=(bytes, str), symlink=(bytes, str))
+    def get_object_symlink(self, bucket_name, symlink, config=None):
+        """
+        Get symlink info
+
+        :type bucket: string
+        :param bucket: None
+
+        :type symlink: string
+        :param symlink: symlink
+
+        :return:
+            **HttpResponse Class**
+        """
+        key = compat.convert_to_bytes(symlink)
+        return self._send_request(
+                http_methods.GET,
+                bucket_name,
+                key,
+                params={b'symlink': b''},
+                config=config)
+
+    @required(bucket_name=(bytes, str), key=(bytes, str), select_object_args=(dict, ))
+    def select_object(self, bucket_name, key, select_object_args, headers=None, config=None):
+        """
+
+        :type bucket_name: string
+        :param bucket_name: bucket name
+
+        :type key: string
+        :param key: object name
+
+        :type select_object_args: dict
+        :param select_object_args: requesta parameters for select object api
+
+        :param config:
+        :return:
+        """
+        key = compat.convert_to_bytes(key)
+        headers = headers or {}
+        if "inputSerialization" in select_object_args and "json" in select_object_args["inputSerialization"]:
+            select_type = b"json"
+        else:
+            select_type = b"csv"
+        select_response = SelectResponse()
+        self._send_request(
+            http_methods.POST,
+            bucket_name,
+            key,
+            body=json.dumps({'selectRequest': select_object_args}, default=BosClient._dump_acl_object),
+            headers=headers,
+            params={b'select': b'', b'type': select_type},
+            config=config,
+            body_parser=lambda http_response, response: BosClient._parse_select_message(
+                http_response, response, select_response)
+            )
+        return select_response
+
     @staticmethod
     def _prepare_object_headers(
             content_length=None,
@@ -1879,7 +2182,10 @@ class BosClient(BceBaseClient):
             etag=None,
             user_metadata=None,
             storage_class=None,
-            user_headers=None):
+            user_headers=None,
+            encryption=None,
+            customer_key=None,
+            customer_key_md5=None):
         headers = {}
 
         if content_length is not None:
@@ -1919,6 +2225,17 @@ class BosClient(BceBaseClient):
         if storage_class is not None:
             headers[http_headers.BOS_STORAGE_CLASS] = storage_class
 
+        if encryption is not None:
+            headers[http_headers.BOS_SERVER_SIDE_ENCRYPTION] = utils.convert_to_standard_string(encryption)
+
+        if customer_key is not None:
+            headers[http_headers.BOS_SERVER_SIDE_ENCRYPTION_CUSTOMER_KEY] = \
+                utils.convert_to_standard_string(customer_key)
+
+        if customer_key_md5 is not None:
+            headers[http_headers.BOS_SERVER_SIDE_ENCRYPTION_CUSTOMER_KEY_MD5] = \
+                utils.convert_to_standard_string(customer_key_md5)
+
         if user_headers is not None:
             try:
                 headers = BosClient._get_user_header(headers, user_headers, False)
@@ -1960,7 +2277,12 @@ class BosClient(BceBaseClient):
 
 
     @staticmethod
-    def _get_path(config, bucket_name=None, key=None):
+    def _get_path(config, bucket_name=None, key=None, use_backup_endpoint=False):
+        host = config.endpoint
+        if use_backup_endpoint:
+            host = config.backup_endpoint
+        if config.cname_enabled or utils.is_cname_like_host(host):
+            return utils.append_uri(bos.URL_PREFIX, key)
         return utils.append_uri(bos.URL_PREFIX, bucket_name, key)
 
     def _merge_config(self, config):
@@ -1971,6 +2293,22 @@ class BosClient(BceBaseClient):
             new_config.merge_non_none_values(config)
             return new_config
 
+    @staticmethod
+    def _need_retry_backup_endpoint(error):
+        # always retry on IOError
+        if isinstance(error, IOError):
+            return True
+
+        # Only retry on a subset of service exceptions
+        if isinstance(error, BceServerError):
+            if error.status_code == http.client.INTERNAL_SERVER_ERROR:
+                return True
+            if error.status_code == http.client.SERVICE_UNAVAILABLE:
+                return True
+            if error.code == BceServerError.REQUEST_EXPIRED:
+                return True
+        return False
+
     def _send_request(
             self, http_method, bucket_name=None, key=None,
             body=None, headers=None, params=None,
@@ -1985,6 +2323,142 @@ class BosClient(BceBaseClient):
             headers = headers or {}
             headers[http_headers.STS_SECURITY_TOKEN] = config.security_token
 
-        return bce_http_client.send_request(
-            config, bce_v1_signer.sign, [handler.parse_error, body_parser],
-            http_method, path, body, headers, params)
+        try:
+            return bce_http_client.send_request(
+                config, bce_v1_signer.sign, [handler.parse_error, body_parser],
+                http_method, path, body, headers, params)
+        except BceHttpClientError as e:
+            # retry backup endpoint
+            if config.backup_endpoint is None:
+                raise e
+            if BosClient._need_retry_backup_endpoint(e.last_error):
+                _logger.debug(b'Retry for backup endpoint.')
+                path = BosClient._get_path(config, bucket_name, key, True)
+                return bce_http_client.send_request(
+                    config, bce_v1_signer.sign, [handler.parse_error, body_parser],
+                    http_method, path, body, headers, params, True)
+            else:
+                raise e
+
+
+class SelectMessage(object):
+    """
+    returned message from select object api
+    """
+    def set_record_message(self, headers, payload, crc):
+        """
+        Initialize for record message
+        """
+        self.type = "Records"
+        self.headers = headers
+        self.payload = payload
+        self.crc = crc
+
+    def set_cont_message(self, headers, bytes_scanned, bytes_returned, crc):
+        """
+        Initialize for continue message
+        """
+        self.type = "Cont"
+        self.headers = headers
+        self.bytes_scanned = bytes_scanned
+        self.bytes_returned = bytes_returned
+        self.crc = crc
+
+    def set_end_message(self, headers, crc):
+        """
+        Initialize for end message
+        """
+        self.type = "End"
+        self.headers = headers
+        self.crc = crc
+
+    def __str__(self):
+        if self.type == "Records":
+            return '{}\n{}'.format(self.headers, self.payload)
+        elif self.type == "Cont":
+            return '{}\nbytes_scanned/bytes_returned={}/{}'.format(self.headers, self.bytes_scanned,
+                                                               self.bytes_returned)
+        else:
+            return '{}'.format(self.headers)
+
+class SelectResponse(object):
+    """
+    deal with message of select object api
+    """
+    def __init__(self):
+        self.finish = False
+
+    def init_from_http_response(self, http_response, response):
+        """
+        get HttpResponse and BceResponse
+        """
+        self.http_response = http_response
+        self.response = response
+
+    def result(self):
+        """
+        generator for SelectMessage
+        """
+        f  = self.http_response
+        try:
+            while not self.finish:
+                prelude = f.read(8)
+                if not prelude:
+                    raise StopIteration
+                    return
+                total_len = struct.unpack('>I', prelude[0:4])[0]
+                headers_len = struct.unpack('>I', prelude[4:8])[0]
+                headers = f.read(headers_len)
+                headers_map = self._parse_select_headers(headers)
+                msg = SelectMessage()
+                if headers_map['message-type'] == 'Records':
+                    payload_len = total_len - headers_len - 12
+                    payload = f.read(payload_len)
+                    crc = struct.unpack('>I', f.read(4))[0]
+                    msg.set_record_message(headers_map, compat.convert_to_string(payload), crc)
+                    yield msg
+                elif headers_map['message-type'] == 'Cont':
+                    bytes_scanned = f.read(8)
+                    bytes_returned = f.read(8)
+                    crc = struct.unpack('>I', f.read(4))[0]
+                    bytes_scanned = struct.unpack('>Q', bytes_scanned)[0]
+                    bytes_returned = struct.unpack('>Q', bytes_returned)[0]
+                    msg.set_cont_message(headers_map, bytes_scanned, bytes_returned, crc)
+                    yield msg
+                elif headers_map['message-type'] == 'End':
+                    crc = struct.unpack('>I', f.read(4))[0]
+                    if headers_map["error-code"] != "success":
+                        raise BceServerError(headers_map['error-message'], code=headers_map['error-code'],
+                                request_id=self.response.metadata.bce_request_id)
+                        return
+                    msg.set_end_message(headers_map, crc)
+                    self.finish = True
+                    yield msg
+            raise StopIteration
+        finally:
+            self.http_response.close()
+
+    @staticmethod
+    def _parse_select_headers(headers):
+        """
+        parse SELECT headers
+        :param headers: <str>
+        :return: <dict>
+        """
+        hm = {}
+        index = 0
+        while index < len(headers):
+            # headers key length
+            key_len = struct.unpack('B', headers[index: index + 1])[0]
+            index += 1
+            # headers key
+            key = headers[index: index + key_len]
+            index += key_len
+            # headers value length
+            value_len = struct.unpack('>H', headers[index: index + 2])[0]
+            index += 2
+            # headers value
+            value = headers[index: index + value_len]
+            index += value_len
+            hm[compat.convert_to_string(key)] = compat.convert_to_string(value)
+        return hm
